@@ -7,14 +7,23 @@ use std::ptr;
 use std::rc::Rc;
 use std::rc::Weak;
 
+use smallvec::SmallVec;
+
 use super::*;
 
+/// The number of effects that are allocated on the stack before resorting to heap allocation in
+/// [`ReactiveScope`].
+const REACTIVE_SCOPE_EFFECTS_STACK_CAPACITY: usize = 4;
+
+/// Initial capacity for [`CONTEXTS`].
+const CONTEXTS_INITIAL_CAPACITY: usize = 10;
+
 thread_local! {
     /// Context of the effect that is currently running. `None` if no effect is running.
     ///
     /// This is an array of callbacks that, when called, will add the a `Signal` to the `handle` in the argument.
     /// The callbacks return another callback which will unsubscribe the `handle` from the `Signal`.
-    pub(super) static CONTEXTS: RefCell<Vec<Weak<RefCell<Option<Running>>>>> = RefCell::new(Vec::new());
+    pub(super) static CONTEXTS: RefCell<Vec<Weak<RefCell<Option<Running>>>>> = RefCell::new(Vec::with_capacity(CONTEXTS_INITIAL_CAPACITY));
     pub(super) static SCOPE: RefCell<Option<ReactiveScope>> = RefCell::new(None);
 }
 
@@ -22,7 +31,7 @@ thread_local! {
 /// When the state is dropped, all dependencies are removed (both links and backlinks).
 pub(super) struct Running {
     /// Callback to run when the effect is recreated.
-    pub(super) execute: Rc<dyn Fn()>,
+    pub(super) execute: Rc<RefCell<dyn FnMut()>>,
     /// A list of dependencies which trigger the effect.
     pub(super) dependencies: HashSet<Dependency>,
     /// The reactive scope owns all effects created within it.
@@ -42,19 +51,13 @@ impl Running {
     }
 }
 
-impl Drop for Running {
-    fn drop(&mut self) {
-        self.clear_dependencies();
-    }
-}
-
 /// Owns the effects created in the current reactive scope.
 /// The effects are dropped and the cleanup callbacks are called when the [`ReactiveScope`] is
 /// dropped.
 #[derive(Default)]
 pub struct ReactiveScope {
     /// Effects created in this scope.
-    effects: Vec<Rc<RefCell<Option<Running>>>>,
+    effects: SmallVec<[Rc<RefCell<Option<Running>>>; REACTIVE_SCOPE_EFFECTS_STACK_CAPACITY]>,
     /// Callbacks to call when the scope is dropped.
     cleanup: Vec<Box<dyn FnOnce()>>,
 }
@@ -91,17 +94,17 @@ impl Drop for ReactiveScope {
 }
 
 #[derive(Clone)]
-pub(super) struct Callback(pub(super) Weak<dyn Fn()>);
+pub(super) struct Callback(pub(super) Weak<RefCell<dyn FnMut()>>);
 
 impl Callback {
     #[track_caller]
     #[must_use = "returned value must be manually called"]
-    pub fn callback(&self) -> Rc<dyn Fn()> {
+    pub fn callback(&self) -> Rc<RefCell<dyn FnMut()>> {
         self.try_callback().expect("callback is not valid anymore")
     }
 
     #[must_use = "returned value must be manually called"]
-    pub fn try_callback(&self) -> Option<Rc<dyn Fn()>> {
+    pub fn try_callback(&self) -> Option<Rc<RefCell<dyn FnMut()>>> {
         self.0.upgrade()
     }
 }
@@ -158,14 +161,13 @@ pub fn create_effect_initial<R: 'static>(
     fn internal(initial: Box<InitialFn>) -> Box<dyn Any> {
         let running: Rc<RefCell<Option<Running>>> = Rc::new(RefCell::new(None));
 
-        type MutEffect = Rc<RefCell<Option<Box<dyn FnMut()>>>>;
-        let effect: MutEffect = Rc::new(RefCell::new(None));
+        let mut effect: Option<Box<dyn FnMut()>> = None;
         let ret: Rc<RefCell<Option<Box<dyn Any>>>> = Rc::new(RefCell::new(None));
 
-        let initial = RefCell::new(Some(initial));
+        let mut initial = Some(initial);
 
         // Callback for when the effect's dependencies are triggered.
-        let execute: Rc<dyn Fn()> = Rc::new({
+        let execute: Rc<RefCell<dyn FnMut()>> = Rc::new(RefCell::new({
             let running = Rc::downgrade(&running);
             let ret = Rc::downgrade(&ret);
             move || {
@@ -177,9 +179,6 @@ pub fn create_effect_initial<R: 'static>(
                     // the effect.
                     let running = running.upgrade().unwrap();
 
-                    // Recreate effect dependencies each time effect is called.
-                    running.borrow_mut().as_mut().unwrap().clear_dependencies();
-
                     // Push new reactive scope.
                     contexts.borrow_mut().push(Rc::downgrade(&running));
 
@@ -189,11 +188,14 @@ pub fn create_effect_initial<R: 'static>(
                         let scope = create_root(|| {
                             // Run initial effect closure.
                             let (effect_tmp, ret_tmp) = initial();
-                            *effect.borrow_mut() = Some(effect_tmp);
+                            effect = Some(effect_tmp);
                             *ret.borrow_mut() = Some(ret_tmp);
                         });
                         running.borrow_mut().as_mut().unwrap().scope = scope;
                     } else {
+                        // Recreate effect dependencies each time effect is called.
+                        running.borrow_mut().as_mut().unwrap().clear_dependencies();
+
                         // Destroy old effects before new ones run.
 
                         // We want to destroy the old scope before creating the new one, so that
@@ -203,7 +205,7 @@ pub fn create_effect_initial<R: 'static>(
 
                         // Run effect closure.
                         let new_scope = create_root(|| {
-                            effect.borrow_mut().as_mut().unwrap()();
+                            effect.as_mut().unwrap()();
                         });
                         running.borrow_mut().as_mut().unwrap().scope = new_scope;
                     }
@@ -227,7 +229,7 @@ pub fn create_effect_initial<R: 'static>(
                     );
                 });
             }
-        });
+        }));
 
         *running.borrow_mut() = Some(Running {
             execute: Rc::clone(&execute),
@@ -256,7 +258,7 @@ pub fn create_effect_initial<R: 'static>(
             }
         });
 
-        execute();
+        execute.borrow_mut()();
 
         let ret = Rc::try_unwrap(ret).unwrap(); // ret should only have 1 strong reference
         ret.into_inner().unwrap()
